use core::time::Duration;

use log::{error, info};
use tokio::time::delay_for;

use crate::connectors::{Client, ExchangeKind};
use crate::currency::{Symbol, SymbolPair};
use crate::events::Event;
use crate::managers::{OrderManager, PositionManager, PriceManager};
use crate::ticker::Ticker;
use crate::BoxError;

pub struct BfxBot {
    ticker: Ticker,
    quote: Symbol,
    trading_symbols: Vec<Symbol>,
    price_managers: Vec<PriceManager>,
    order_managers: Vec<OrderManager>,
    pos_managers: Vec<PositionManager>,
}

impl BfxBot {
    pub fn new(
        exchanges: Vec<ExchangeKind>,
        trading_symbols: Vec<Symbol>,
        quote: Symbol,
        tick_duration: Duration,
    ) -> Self {
        let clients: Vec<_> = exchanges.iter().map(|x| Client::new(x)).collect();
        let pairs: Vec<_> = trading_symbols
            .iter()
            .map(|x| SymbolPair::new(quote.clone(), x.clone()))
            .collect();

        let mut pos_managers = Vec::new();
        let mut order_managers = Vec::new();
        let mut pair_statuses = Vec::new();

        for c in clients {
            for p in &pairs {
                pos_managers.push(PositionManager::new(p.clone(), c.clone()));
                order_managers.push(OrderManager::new(p.clone(), c.clone()));
                pair_statuses.push(PriceManager::new(p.clone(), c.clone()));
            }
        }

        BfxBot {
            ticker: Ticker::new(tick_duration),
            price_managers: pair_statuses,
            quote,
            trading_symbols,
            order_managers,
            pos_managers,
        }
    }

    pub async fn start_loop(&mut self) -> Result<(), BoxError> {
        if let Err(e) = self.update_managers().await {
            error!("Error while starting managers: {}", e);
        }

        loop {
            info!("Current tick: {}", self.ticker.current_tick());

            if let Err(e) = self.update().await {
                error!("Error in main bot loop: {}", e);
            }
        }
    }

    async fn update(&mut self) -> Result<(), BoxError> {
        delay_for(self.ticker.duration()).await;
        self.ticker.inc();

        if let Err(e) = self.update_managers().await {
            error!("Error while updating managers: {}", e);
        }

        Ok(())
    }

    async fn update_managers(&mut self) -> Result<(), BoxError> {
        self.update_price_managers().await?;
        self.update_position_managers().await?;

        Ok(())
    }

    async fn update_position_managers(&mut self) -> Result<Option<Vec<Event>>, BoxError> {
        for mgr in &mut self.pos_managers {
            let tick = self.ticker.current_tick();

            mgr.update(tick).await?;
        }

        Ok(None)
    }

    async fn update_price_managers(&mut self) -> Result<Option<Vec<Event>>, BoxError> {
        let futures: Vec<_> = self
            .price_managers
            .clone()
            .into_iter()
            // the only reason you need the async block is that the future
            // returned by x.update(tick) borrows from x
            // so we create a future that first takes ownership of x, then uses it to call x.update
            .map(|mut x| {
                let tick = self.ticker.current_tick();
                async move { x.update(tick).await }
            })
            .map(tokio::spawn)
            .collect();

        let mut price_entries = vec![];

        for f in futures {
            price_entries.push(f.await??);
        }

        for manager in &mut self.price_managers {
            let prices: Vec<_> = price_entries
                .drain_filter(|x| x.pair() == manager.pair())
                .collect();

            for p in prices {
                manager.add_entry(p);
            }
        }

        Ok(None)
    }
}
